"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = void 0;
var _mapValues = _interopRequireDefault(require("lodash/mapValues"));
var _constant = _interopRequireDefault(require("lodash/constant"));
var _cloneDeep = _interopRequireDefault(require("lodash/cloneDeep"));
var _sum = _interopRequireDefault(require("lodash/sum"));
var _values = _interopRequireDefault(require("lodash/values"));
var _keys = _interopRequireDefault(require("lodash/keys"));
var _reduce = _interopRequireDefault(require("lodash/reduce"));
var _forEach = _interopRequireDefault(require("lodash/forEach"));
var _isNumber = _interopRequireDefault(require("lodash/isNumber"));
var _clamp = _interopRequireDefault(require("lodash/clamp"));
var _has = _interopRequireDefault(require("lodash/has"));
var _isUndefined = _interopRequireDefault(require("lodash/isUndefined"));
var _isObject = _interopRequireDefault(require("lodash/isObject"));
var _isArray = _interopRequireDefault(require("lodash/isArray"));
var _includes = _interopRequireDefault(require("lodash/includes"));
var _pick = _interopRequireDefault(require("lodash/pick"));
var _without = _interopRequireDefault(require("lodash/without"));
var _isEmpty = _interopRequireDefault(require("lodash/isEmpty"));
var _max = _interopRequireDefault(require("lodash/max"));
var _findKey = _interopRequireDefault(require("lodash/findKey"));
var _omit = _interopRequireDefault(require("lodash/omit"));
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }
function _toConsumableArray(arr) { return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _unsupportedIterableToArray(arr) || _nonIterableSpread(); }
function _nonIterableSpread() { throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); }
function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }
function _iterableToArray(iter) { if (typeof Symbol !== "undefined" && iter[Symbol.iterator] != null || iter["@@iterator"] != null) return Array.from(iter); }
function _arrayWithoutHoles(arr) { if (Array.isArray(arr)) return _arrayLikeToArray(arr); }
function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }
function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }
function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }
function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); Object.defineProperty(Constructor, "prototype", { writable: false }); return Constructor; }
function _classPrivateMethodInitSpec(obj, privateSet) { _checkPrivateRedeclaration(obj, privateSet); privateSet.add(obj); }
function _checkPrivateRedeclaration(obj, privateCollection) { if (privateCollection.has(obj)) { throw new TypeError("Cannot initialize the same private elements twice on an object"); } }
function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }
function _classPrivateMethodGet(receiver, privateSet, fn) { if (!privateSet.has(receiver)) { throw new TypeError("attempted to get private field on non-instance"); } return fn; }
var _ = {
  mapValues: _mapValues["default"],
  forEach: _forEach["default"],
  constant: _constant["default"],
  cloneDeep: _cloneDeep["default"],
  sum: _sum["default"],
  values: _values["default"],
  keys: _keys["default"],
  reduce: _reduce["default"],
  isNumber: _isNumber["default"],
  clamp: _clamp["default"],
  has: _has["default"],
  isUndefined: _isUndefined["default"],
  isObject: _isObject["default"],
  isArray: _isArray["default"],
  includes: _includes["default"],
  pick: _pick["default"],
  without: _without["default"],
  isEmpty: _isEmpty["default"],
  max: _max["default"],
  findKey: _findKey["default"],
  omit: _omit["default"]
};

/**
 * const roller = new Brng(config)
 *
 * Constructor parameters:
 *  originalProportions [REQUIRED] {Object[String/Number:Number]} -- key-value mapping of
 *    weighted proportions. For example {mickeyd: 3, jackinthebox: 3, burgerking: 2, whataburger: 10}
 *  config {Object}
 *  config.random {Function} -- function that returns random number 0 - 1. Defaults to Math.random
 *  config.keepHistory {Boolean} -- if true, keep the roll history
 *  config.bias {Number} -- between 0 and 4. The higher the bias, the more it
 *    favors values less chosen. If 0, it's basically a normal RNG. Defaults to 1.
 *  config.repeatTolerance {Number} -- between 0 and 1. The lower the tolerance, the more
 *    likely Brng will re-roll if it's a repeat. If 0, it'll never repeat. Defaults to 1.
 *
 * Public methods:
 *  roll() -- selects a random value; remembers previous rolls.
 *  roll(value) -- force select the value from your original proportions. Ignores all criteria.
 *  roll({exclude: [value1, value2]}) -- select a value that excludes any values in the array
 *  roll({only: [value1, value2]}) -- select one of the values only within the given array
 * 
 *  flip(), pick(), select(), choose(), randomize() -- aliases of `roll()`
 *  reset() -- resets all history and resets previous rolls
 *  undo () -- undo the previous roll. must have `config.keepHistory === true`
 *
 * Public readable values (if keepHistory === true):
 *  historyArray -- array of every single previous roll
 *  historyMapping -- hash of key to number of times rolled in total
 *
 * Public static values for common default proportions:
 *  Brng.defaultProportions.one6SidedDie -- proportions for rolling 1 6-sided die
 *  Brng.defaultProportions.two6SidedDice -- proportions for rolling 2 6-sided dice
 *  Brng.defaultProportions.coin -- proportions for flipping 1 coin
 */
var _setupFromOriginalProportions = /*#__PURE__*/new WeakSet();
var _shiftProportions = /*#__PURE__*/new WeakSet();
var _reverseShiftProportions = /*#__PURE__*/new WeakSet();
var _rollSideEffects = /*#__PURE__*/new WeakSet();
var Brng = /*#__PURE__*/function () {
  function Brng(_originalProportions) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    _classCallCheck(this, Brng);
    _classPrivateMethodInitSpec(this, _rollSideEffects);
    _classPrivateMethodInitSpec(this, _reverseShiftProportions);
    _classPrivateMethodInitSpec(this, _shiftProportions);
    _classPrivateMethodInitSpec(this, _setupFromOriginalProportions);
    _defineProperty(this, "flip", this.roll);
    _defineProperty(this, "pick", this.roll);
    _defineProperty(this, "select", this.roll);
    _defineProperty(this, "choose", this.roll);
    _defineProperty(this, "randomize", this.roll);
    _defineProperty(this, "update", this.add);
    this.keepHistory = !!config.keepHistory;
    if (this.keepHistory) {
      this.historyMapping = _.mapValues(_originalProportions, _.constant(0));
      this.historyArray = [];
    }
    this.biasMultiplier = _.isNumber(config.bias) ? _.clamp(config.bias, 0, 4) : 1;
    this.previousRoll = null;
    this.repeatTolerance = _.isNumber(config.repeatTolerance) ? _.clamp(config.repeatTolerance, 0, 1) : 1;
    this.random = config.random || Math.random;
    _classPrivateMethodGet(this, _setupFromOriginalProportions, _setupFromOriginalProportions2).call(this, _originalProportions);

    // set up for the rest of the time
    this.proportions = _.cloneDeep(_originalProportions);
  }
  _createClass(Brng, [{
    key: "chooseKeyFromProportions",
    value:
    // The first step of the algorithm: given a weighted proportion map, randomly select a value
    function chooseKeyFromProportions() {
      var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
      // optional config

      var availableKeys = this.possibleKeys;
      if (_.isArray(config.only)) {
        availableKeys = config.only;
      }
      if (_.isArray(config.exclude)) {
        availableKeys = _.without.apply(_, [availableKeys].concat(_toConsumableArray(config.exclude)));
      }
      if (_.isEmpty(availableKeys)) {
        throw new Error('The values given in `only` or `exclude` gave no available options.');
      }
      var availableProportions = _.pick(this.proportions, availableKeys);
      var valuesOfAvailableOptions = _.values(availableProportions);
      var sumTotal = _.sum(valuesOfAvailableOptions);

      // RETURNS the key with the highest proportion if the sum total is negative
      if (sumTotal <= 0) {
        var highestProportion = _.max(valuesOfAvailableOptions);
        var chosenKey = _.findKey(availableProportions, function (val) {
          return val === highestProportion;
        });
        return chosenKey;
      }

      // ELSE if everything is normal, proceed as normal
      var rawRoll = this.random() * sumTotal;
      var currentSum = 0;
      for (var i = 0; i < availableKeys.length; i++) {
        var _chosenKey = availableKeys[i];
        var proportionValue = this.proportions[_chosenKey];
        currentSum += proportionValue;
        if (currentSum > rawRoll) {
          return _chosenKey;
        }
      }
      throw new Error('Something is wrong. Code should not reach here.');
    }

    // The second step of the algorithm: given the selected value, shift the proportion map around that value
    // WARNING: method writes to this.proportions
  }, {
    key: "passCriteria",
    value: function passCriteria(keyChosen) {
      if (this.previousRoll === keyChosen && this.repeatTolerance < 1) {
        return this.random() < this.repeatTolerance;
      } else {
        return true;
      }
    }
  }, {
    key: "roll",
    value: function /* value or config */
    roll() {
      // optional
      var keyFromArgs = arguments[0];
      var keyIsAvailable = _.has(this.originalProportions, keyFromArgs);
      var keyChosen;
      if (!_.isUndefined(keyFromArgs) && keyIsAvailable) {
        // this.roll(value) is called
        keyChosen = keyFromArgs;
      } else if (!_.isUndefined(keyFromArgs) && !keyIsAvailable && !_.isObject(keyFromArgs)) {
        // this.roll(value) is called incorectly
        throw new Error('The value ' + keyFromArgs + ' is not in your originalProportions.');
      } else {
        // the normal this.roll() with a possible config
        var config = arguments[0];
        keyChosen = this.chooseKeyFromProportions(config);
        if (!this.passCriteria(keyChosen)) {
          return this.roll();
        }
      }
      _classPrivateMethodGet(this, _shiftProportions, _shiftProportions2).call(this, keyChosen);
      _classPrivateMethodGet(this, _rollSideEffects, _rollSideEffects2).call(this, keyChosen);
      return keyChosen;
    }
  }, {
    key: "undo",
    value: function undo() {
      if (!this.keepHistory) {
        // throw if there's no keepHistory
        throw new Error('To undo, set `{keepHistory: true}`.');
      }
      var previousChoice = this.historyArray.pop(); // ! writes to historyArray
      this.historyMapping[previousChoice] = this.historyMapping[previousChoice] - 1;
      _classPrivateMethodGet(this, _reverseShiftProportions, _reverseShiftProportions2).call(this, previousChoice);
      return previousChoice;
    }
  }, {
    key: "reset",
    value: function reset() {
      this.proportions = _.cloneDeep(this.originalProportions);
      if (this.keepHistory) {
        this.historyMapping = _.mapValues(this.originalProportions, _.constant(0));
        this.historyArray = [];
      }
    }

    // roll() won't call the paused key anymore, and the proportions won't be affected
  }, {
    key: "pause",
    value: function pause(key) {}
  }, {
    key: "unpause",
    value: function unpause(key) {}

    // add the key to the possible proportinos
  }, {
    key: "add",
    value: function add(newProportions) {
      var _this = this;
      var originalProportions = _.cloneDeep(this.originalProportions);
      _.forEach(newProportions, function (proportionWeight, key) {
        originalProportions[key] = proportionWeight;
        if (!_.has(_this.proportions, key)) {
          _this.proportions[key] = proportionWeight;
        }
      });
      _classPrivateMethodGet(this, _setupFromOriginalProportions, _setupFromOriginalProportions2).call(this, originalProportions);
    }
  }, {
    key: "remove",
    value: function remove(keyToRemove) {
      var newOriginalProportions = _.omit(this.originalProportions, keyToRemove);
      _classPrivateMethodGet(this, _setupFromOriginalProportions, _setupFromOriginalProportions2).call(this, newOriginalProportions);
    }
  }]);
  return Brng;
}();
function _setupFromOriginalProportions2(originalProportions) {
  var _this2 = this;
  this.originalProportions = _.cloneDeep(originalProportions);
  this.possibleKeys = _.keys(originalProportions);
  var sumTotal = _.sum(_.values(originalProportions));
  this.originalProbabilities = _.mapValues(originalProportions, function (number) {
    return number / sumTotal;
  });
  var baseValueToRedistribute = _.reduce(originalProportions, function (sum, proportionValue, keyId) {
    return sum + proportionValue * _this2.originalProbabilities[keyId];
  }, 0);
  this.valueToRedistribute = baseValueToRedistribute * this.biasMultiplier;
}
function _shiftProportions2(keyChosen) {
  var _this3 = this;
  this.proportions[keyChosen] = this.proportions[keyChosen] - this.valueToRedistribute;
  _.forEach(this.possibleKeys, function (otherKey) {
    _this3.proportions[otherKey] = _this3.proportions[otherKey] + _this3.valueToRedistribute * _this3.originalProbabilities[otherKey];
  });
}
function _reverseShiftProportions2(keyChosen) {
  var _this4 = this;
  this.proportions[keyChosen] = this.proportions[keyChosen] + this.valueToRedistribute;
  _.forEach(this.possibleKeys, function (otherKey) {
    _this4.proportions[otherKey] = _this4.proportions[otherKey] - _this4.valueToRedistribute * _this4.originalProbabilities[otherKey];
  });
}
function _rollSideEffects2(keyChosen) {
  this.previousRoll = keyChosen;
  if (this.keepHistory) {
    this.historyMapping[keyChosen] = this.historyMapping[keyChosen] ? this.historyMapping[keyChosen] + 1 : 1;
    this.historyArray.push(keyChosen);
  }
  return keyChosen;
}
_defineProperty(Brng, "defaultProportions", {
  one6SidedDie: {
    1: 1,
    2: 1,
    3: 1,
    4: 1,
    5: 1,
    6: 1
  },
  two6SidedDice: {
    2: 1,
    3: 2,
    4: 3,
    5: 4,
    6: 5,
    7: 6,
    8: 5,
    9: 4,
    10: 3,
    11: 2,
    12: 1
  },
  coin: {
    heads: 1,
    tails: 1
  }
});
var _default = Brng;
exports["default"] = _default;
